/* Author: Tobi Vollebregt */

/*  Marsworm -- A small game written for the TINS Is Not SpeedHack 2003
 *  Copyright (C) 2003  Tobi Vollebregt
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *  See `License.txt', which contains a verbatim copy of the
 *  GNU General Public License, for details.
 *
 *  Please send your reaction to: tobivollebregt@hotmail.com
 */

#include <math.h>
#include <allegro.h>
#include <allegro/internal/aintern.h>
#include "aaline.h"

#define SWAP(a,b)   {temp=a;a=b;b=temp;}

void _linear_aa_line_8(BITMAP *b, int x1, int y1, int x2, int y2, int color)
{
	int xdiff=x2-x1,ydiff=y2-y1;
	int temp,deltax,deltay,x,y,yc,c;
	unsigned long address;

	if(xdiff==0 && ydiff==0) return;

	/*if((x1<0 && x2<0) || (x1>=b->w && x2>=b->w)) return;
	if((y1<0 && y2<0) || (y1>=b->h && y2>=b->h)) return;*/

	bmp_select(b);

	if(abs(xdiff)>abs(ydiff))
	{
		if(x1>x2) {SWAP(x1,x2);SWAP(y1,y2);}

		deltay=65536*ydiff/xdiff;

		/*if(x1<0) {y1-=(x1*deltay)>>16;x1=0;}
		if(x2>b->w-1) {y2-=((x2-(b->w-1))*deltay)>>16;x2=b->w-1;}
		if(y1<0) {x1-=65536*y1/deltay;y1=0;}

		if(y1>b->h-2) {x1+=(y1-(b->h-2))/deltay;y1=b->h-2;}

		if(y2<0) {x2-=65536*y2/deltay;y2=0;}

		if(y2>b->h-2) {x2+=(y2-(b->h-2))/deltay;y2=b->h-2;}

		if(x1==x2 && y1==y2) return;*/

		for(x=x1,y=65536*y1; x<x2; x++,y+=deltay)
		{
			yc=y>>16;

			address=(unsigned long)b->line[yc]+x;
			c=bmp_read8(address)+63-((y>>10)&0x3F);
			bmp_write8(address,MIN(c,63));

			address=(unsigned long)b->line[yc+1]+x;
			c=bmp_read8(address)+((y>>10)&0x3F);
			bmp_write8(address,MIN(c,63));
		}
	}
	else
	{
		if(y1>y2) {SWAP(x1,x2);SWAP(y1,y2);}

		deltax=65536*xdiff/ydiff;

		for(y=y1,x=65536*x1; y<y2; y++,x+=deltax)
		{
			address=(unsigned long)b->line[y]+(x>>16);
			c=bmp_read8(address)+63-((x>>10)&0x3F);
			bmp_write8(address,MIN(c,63));

			address++;
			c=bmp_read8(address)+((x>>10)&0x3F);
			bmp_write8(address,MIN(c,63));
		}
	}
}

#define LINEAR_AA_LINE(depth,bytes) \
					\
void _linear_aa_line_##depth(BITMAP *b, int x1, int y1, int x2, int y2, int color)  \
{                                   \
	int xdiff=x2-x1,ydiff=y2-y1;    \
	int temp,deltax,deltay,x,y,yc,a;    \
	unsigned long address;          \
					\
	if(xdiff==0 && ydiff==0) return;    \
					\
	bmp_select(b);                  \
					\
	if(abs(xdiff)>abs(ydiff))       \
	{                               \
		if(x1>x2) {SWAP(x1,x2);SWAP(y1,y2);}    \
						\
		deltay=65536*ydiff/xdiff;   \
						\
		for(x=x1,y=65536*y1; x<x2; x++,y+=deltay)   \
		{                           \
			yc=y>>16;               \
						\
			a=(y>>8)&0xFF;          \
						\
			address=(unsigned long)b->line[yc]+bytes*x; \
			bmp_write##depth(address,_blender_trans##depth(color,bmp_read##depth(address),255-a));  \
						\
			yc++;                   \
			address=(unsigned long)b->line[yc]+bytes*x; \
			bmp_write##depth(address,_blender_trans##depth(color,bmp_read##depth(address),a));  \
		}                           \
	}                               \
	else                            \
	{                               \
		if(y1>y2) {SWAP(x1,x2);SWAP(y1,y2);}    \
						\
		deltax=65536*xdiff/ydiff;   \
						\
		for(y=y1,x=65536*x1; y<y2; y++,x+=deltax)   \
		{                           \
			a=(x>>8)&0xFF;          \
						\
			address=(unsigned long)b->line[y]+bytes*(x>>16);    \
			bmp_write##depth(address,_blender_trans##depth(color,bmp_read##depth(address),255-a));  \
						\
			address+=bytes;         \
			bmp_write##depth(address,_blender_trans##depth(color,bmp_read##depth(address),a));  \
		}                           \
	}                               \
}                                   \

#define _blender_trans32    _blender_trans24

LINEAR_AA_LINE(15,2)
LINEAR_AA_LINE(16,2)
LINEAR_AA_LINE(24,3)
LINEAR_AA_LINE(32,4)

#undef LINEAR_AA_LINE
#undef _blender_trans32

void aa_line(BITMAP *b, int x1, int y1, int x2, int y2, int color)
{
	if(y1==y2) hline(b,x1,y1,x2,color);
	if(x1==x2) vline(b,x1,y1,y2,color);

	switch(bitmap_color_depth(b))
	{
		case 8:  _linear_aa_line_8 (b,x1,y1,x2,y2,color); break;
		case 15: _linear_aa_line_15(b,x1,y1,x2,y2,color); break;
		case 16: _linear_aa_line_16(b,x1,y1,x2,y2,color); break;
		case 24: _linear_aa_line_24(b,x1,y1,x2,y2,color); break;
		case 32: _linear_aa_line_32(b,x1,y1,x2,y2,color); break;
	}
}
