/* Author: Tobi Vollebregt */

/*  Marsworm -- A small game written for the TINS Is Not SpeedHack 2003
 *  Copyright (C) 2003  Tobi Vollebregt
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *  See `License.txt', which contains a verbatim copy of the
 *  GNU General Public License, for details.
 *
 *  Please send your reaction to: tobivollebregt@hotmail.com
 */

#include <allegro.h>
#include <string.h>
#include "moon.h"
#include "main.h"
#include "monster.h"

static char* text[] =
{
    "N2",
    "O2",
    "Ar",
    "CO2",
    "Ne",
    "He",
    "CH4",
    "Kr",
    "NO",
    "H2",
    "Xe",
    "H2O",
    "O3",
    "O",
    "N",
    "H"
};

Monsters::~Monsters()
{
    for (int i= 0; i<32; ++i)
        destroy_bitmap(blob[i]);
    destroy_bitmap(eye);
    for (int j = 0; j<MAX_FOOD; ++j)
        destroy_bitmap(food[j].food_bmp);
}

void Monsters::initialize()
{
    int i;
    for (i=0; i<32; ++i)
    {
        blob[i] = create_bitmap(32-i, 32-i);
        if (!blob[i]) exitmsg("not enough memory");

        if (worm_type)
            clear_to_color(blob[i], makecol(255, 255, 191));
        else
        {
            int xc = blob[i]->w>>1, yc = blob[i]->h>>1;

            for (int y = 0; y<blob[i]->h; ++y)
                for (int x = 0; x < blob[i]->w; ++x)
                {
                    int d = ((x-xc)*(x-xc)+(y-yc)*(y-yc));
                    if (d == 0) d = 1;
                    int c = 4*blob[i]->w*blob[i]->w / d;
                    if (c>255) c = 255;
                    putpixel(blob[i], x, y, makecol(0, c, 0));
                }
        }
        /*stretch_blit(blob[i], screen, 0, 0, blob[i]->w, blob[i]->h, 2*i, 2*i, blob[i]->w*4, blob[i]->h*4);
        readkey();*/
    }
    eye = load_bitmap("eye.bmp", 0);
    if (!eye) exitmsg("%s: %s", "eye.bmp", ustrerror(errno));

    memset((char*) enemies, 0, sizeof(enemies));
    enemies[0].length=5;
    for (i=1; i<enemies[0].length; ++i)
        enemies[0].trail[i].setxyz(0., 0., 1.);
    enemies[0].trail[0].setxyz(0., 0.1, 1.);
    enemies[0].trail[0].normalize();
    enemies[0].trail[0] *= 1.025;
    //hit = false;

    for (i =0; i<MAX_FOOD; ++i)
    {
        food[i].text_index = rand() % (sizeof(text)/sizeof(text[0]));
        int length = text_length(font, text[food[i].text_index]);
        food[i].food_bmp = create_bitmap(length, 12);
        if (!food[i].food_bmp) exitmsg("not enough memory");
        clear_to_color(food[i].food_bmp, makecol(255, 0, 255));
        food[i].pos.setPolar((float) rand() / RAND_MAX * 2*PI, (float) rand() / RAND_MAX * PI - PI/2, 1.025);
        const char* s = text[food[i].text_index];
        int c = 0;
        for (; *s; ++s, ++c)
        {
            char tmp[2];
            tmp[0]=*s;
            tmp[1]=0;
            text_mode(-1);
            textout(food[i].food_bmp, font, tmp, 8*c, uisdigit(*s) ? 4 : 0, makecol(255, 255, 255));
        }
    }
    score = 0;
    dir0=0.;
}

int Monsters::update(const Vector& player, bool change)
{
    for (int i=0; i<MAX_ENEMIES; ++i) if (enemies[i].length > 0)
    {
        int j;
        if ((enemies[i].trail[0]-enemies[i].trail[1]).magnitude() > 0.03)
            for (j=enemies[i].length-1; j>=1; --j)
            {
                //Vector dir = enemies[i].trail[j-1] - enemies[i].trail[j];
                //float mag = dir.magnitude();
                //dir *= 0.1;
                //dir *= (1./dir.magnitude()) * 0.0001;
                //enemies[i].trail[j] += dir;
                //enemies[i].trail[j].normalize();
                //enemies[i].trail[j] *= 1.025;
                enemies[i].trail[j] = enemies[i].trail[j-1];
            }

        if (change)
        {
            //if (memcmp((char*) &player, (char*) &old_player, sizeof(Vector)))
            Vector dir = player - enemies[i].trail[0];
            //dir.normalize();
            dir0=dir;
        }
        float dot = dir0 * enemies[i].trail[0];
        dir0 -= dot * enemies[i].trail[0];
        dir0 *= (1./dir0.magnitude()) * 0.005;

        enemies[i].trail[0] += dir0;
        enemies[i].trail[0].normalize();
        enemies[i].trail[0] *= 1.025;

        for (j=1; j<enemies[i].length; ++j)
        {
            if (j!=1 && (enemies[i].trail[j] - enemies[i].trail[0]).magnitude() < 0.02)
            {
                return score;
            }

            for (int k = 0; k<MAX_FOOD; ++k)
            {
                if ((enemies[i].trail[j] - food[k].pos).magnitude() < 0.04)
                {
                    int ol = enemies[i].length;
                    enemies[i].length+=food[k].text_index;
                    score += 1+food[k].text_index*food[k].text_index;
                    if (enemies[i].length >= TRAIL_LENGTH) enemies[i].length = TRAIL_LENGTH-1;
                    for (int l = ol; l < enemies[i].length; ++l)
                        enemies[i].trail[l] = enemies[i].trail[ol-1];

                    food[k].text_index = rand() % (sizeof(text)/sizeof(text[0]));
                    int length = text_length(font, text[food[k].text_index]);
                    destroy_bitmap(food[k].food_bmp);
                    food[k].food_bmp = create_bitmap(length, 12);
                    if (!food[k].food_bmp) exitmsg("not enough memory");
                    clear_to_color(food[k].food_bmp, makecol(255, 0, 255));
                    food[k].pos.setPolar((float) rand() / RAND_MAX * 2*PI, (float) rand() / RAND_MAX * PI - PI/2, 1.025);
                    const char* s = text[food[k].text_index];
                    int c = 0;
                    for (; *s; ++s, ++c)
                    {
                        char tmp[2];
                        tmp[0]=*s;
                        tmp[1]=0;
                        text_mode(-1);
                        textout(food[k].food_bmp, font, tmp, 8*c, uisdigit(*s) ? 4 : 0, makecol(255, 255, 255));
                    }
                }
            }
        }
    }
    /*if ((rand() & 1023) == 0)
    {
        for (int i = 0; i < MAX_ENEMIES; ++i) if (enemies[i].length <= 0)
        {
            enemies[i].length=TRAIL_LENGTH;
            for (int j=0; j<enemies[i].length; ++j)
                enemies[i].trail[j] = -player;
        }
    }*/
    /*bool h = hit;
    if (hit && bl) *bl = blood;
    hit = false;
    return h;*/

    return -1;
}

void Monsters::draw(BITMAP*bmp, MATRIX_f* matrix, Moon* moon)
{
        float fx, fy, fz;
        int ix, iy, n, i;

    //set_add_blender(0,0,0,255);
    for (i =0; i<MAX_ENEMIES; ++i) if (enemies[i].length > 0)
    {
        for (int j=enemies[i].length-1; j>=0; --j)
        {
            apply_matrix_f(matrix, enemies[i].trail[j].x(), enemies[i].trail[j].y(),
                    enemies[i].trail[j].z(), &fx, &fy, &fz);
            if (fz > 0.)
            {
                /*if (fabs(fx) < 0.05 && fabs(fy) < 0.05 && fabs(fz-0.975) < 0.05)
                {
                    hit = true;
                    blood = enemies[i].trail[j];
                    //blood.setxyz(enemies[i].trail[j].x()-fx, enemies[i].trail[j].y()-fy,
                      //      enemies[i].trail[j].z()-fz+0.975);
                }*/

                persp_project_f(fx, fy, fz, &fx, &fy);

                n = 32 - (int) (31 / fz);
                n = MID(0, n, 31);
                ix = (int) fx;
                iy = (int) fy;

                if (fz < 2.)
                    moon->addBitmap(blob[n], ix, iy, (int) (256*fz), 255);
                //draw_trans_sprite(bmp, blob[n], ix, iy);
            }
        }

        int spl = (int) (8. / fz);
        //draw_sprite(bmp, eye, ix-eye->w/2 - spl, iy-eye->h/2);
        //draw_sprite(bmp, eye, ix-eye->w/2 + spl, iy-eye->h/2);
        moon->addBitmap(eye, ix - spl, iy, (int) (256*fz)-1);
        moon->addBitmap(eye, ix + spl, iy, (int) (256*fz)-1);
    }

    for (i = 0; i <MAX_FOOD; ++i)
    {
        apply_matrix_f(matrix, food[i].pos.x(), food[i].pos.y(), food[i].pos.z(), &fx, &fy, &fz);
        persp_project_f(fx, fy, fz, &fx, &fy);

        ix = (int) fx;
        iy = (int) fy;

        moon->addBitmap(food[i].food_bmp, ix, iy, (int) (256*fz));
    }
    text_mode(0);
    textprintf_centre(bmp, font, SCREEN_W/2-1, 0, makecol(255, 255, 255), "score: %d", score);
}
