/* Author: Tobi Vollebregt */

/*  Marsworm -- A small game written for the TINS Is Not SpeedHack 2003
 *  Copyright (C) 2003  Tobi Vollebregt
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *  See `License.txt', which contains a verbatim copy of the
 *  GNU General Public License, for details.
 *
 *  Please send your reaction to: tobivollebregt@hotmail.com
 */

#include <allegro.h>
#include <string.h>
#include "aaline.h"
#include "main.h"
#include "moon.h"
#include "monster.h"

#define FOG_BEGIN (1*256)
#define FOG_END (4*256)
extern "C" void set_my_add_blender(int r,int g,int b,int a);

Moon::~Moon()
{
    delete[] vtx;
    delete[] rotated_vtx;
    delete[] tri;
    delete[] edge_done;
    delete[] zsorter;
    destroy_bitmap(selection_buffer);
    destroy_bitmap(mars);
    /*destroy_bitmap(sorcerer[0]);
    destroy_bitmap(sorcerer[1]);*/
    delete parts;
    delete[] tri_normal;
}

void Moon::initialize(int recur)
{
    int i;

    mars = load_bitmap("mars.bmp", 0);
    if (!mars) exitmsg("%s: %s\n", "mars.bmp", ustrerror(errno));
    /*sorcerer[0] = load_bitmap("sorcerer1.bmp", 0);
    if (!sorcerer[0]) exitmsg("%s: %s\n", "sorcerer1.bmp", ustrerror(errno));
    sorcerer[1] = load_bitmap("sorcerer2.bmp", 0);
    if (!sorcerer[1]) exitmsg("%s: %s\n", "sorcerer2.bmp", ustrerror(errno));*/

    theta = phi = 0.;

    /* Ripped the coordinates and triangle indices of an icosahedron from the
    OpenGL Programming Guide (Addison-Wesley Publishing Company) */
    #define X .525731112119133606 
    #define Z .850650808352039932

    static float vdata[12][3] = {
    {-X, 0.0, Z}, {X, 0.0, Z}, {-X, 0.0, -Z}, {X, 0.0, -Z},
    {0.0, Z, X}, {0.0, Z, -X}, {0.0, -Z, X}, {0.0, -Z, -X},
    {Z, X, 0.0}, {-Z, X, 0.0}, {Z, -X, 0.0}, {-Z, -X, 0.0}};
    static unsigned char tindices[20][3] = {
    {0,4,1}, {0,9,4}, {9,5,4}, {4,5,8}, {4,8,1},    
    {8,10,1}, {8,3,10}, {5,3,8}, {5,2,3}, {2,7,3},    
    {7,10,3}, {7,6,10}, {7,11,6}, {11,0,6}, {0,1,6}, 
    {6,1,10}, {9,0,11}, {9,11,2}, {9,2,5}, {7,2,11}};

    #undef X
    #undef Z

    vtx = new Vector[12];
    for (i = 0; i < 12; ++i) vtx[i] = vdata[i];
    tri = new Triangle[20];
    for (i = 0; i < 20; ++i) tri[i].setIndices(tindices[i]);

    num_vertices = 12;
    num_triangles = 20;

    while (recur--) subdivide();

    tri_normal = new Vector[num_triangles];
    rotated_vtx = new Vector[num_vertices];
    edge_done = new bool[num_vertices * num_vertices];
    zsorter_size = 10000;
    zsorter = new ZSorter[zsorter_size];
    memset((char*) zsorter, 0, sizeof(ZSorter)*zsorter_size);
    //matrix=identity_matrix_f;
    slerp = 99.;
    slerpspeed=0.;
    beginquat = endquat = identity_quat;
    parts = new Particle;
    align.setxyz(0., 0., -1.);
    monsters.initialize();
    change=true;
    wireframe=false;
}

void Moon::subdivide()
{
    Vector* new_vtx = new Vector[num_vertices + num_triangles + (num_triangles >> 1)];
    Triangle* new_tri = new Triangle[4 * num_triangles];

    memcpy((char*)new_vtx, (char*)vtx, sizeof(Vector) * num_vertices);

    Vector* fv = new_vtx + num_vertices;
    Triangle* ft = new_tri;
    Triangle* tend = tri + num_triangles;
    for (Triangle* t = tri; t < tend; ++t)
    {
        Vector v1 = vtx[t->i1] + vtx[t->i2];
        Vector v2 = vtx[t->i2] + vtx[t->i3];
        Vector v3 = vtx[t->i3] + vtx[t->i1];

        v1.normalize();
        v2.normalize();
        v3.normalize();

        int c1 = findVertex(v1, new_vtx + num_vertices, fv);
        int c2 = findVertex(v2, new_vtx + num_vertices, fv);
        int c3 = findVertex(v3, new_vtx + num_vertices, fv);

        if (c1 == -1) { c1 = fv - new_vtx; *fv = v1; ++fv; }
        else c1 += num_vertices;
        if (c2 == -1) { c2 = fv - new_vtx; *fv = v2; ++fv; }
        else c2 += num_vertices;
        if (c3 == -1) { c3 = fv - new_vtx; *fv = v3; ++fv; }
        else c3 += num_vertices;

        ft->setIndices(t->i1, c1, c3); ++ft;
        ft->setIndices(t->i2, c2, c1); ++ft;
        ft->setIndices(t->i3, c3, c2); ++ft;
        ft->setIndices(c1, c2, c3); ++ft;
    }

    TRACE("vertices:%d/%d triangles:%d/%d\n", (int)(fv - new_vtx),
        num_vertices + num_triangles + (num_triangles >> 1), (int)(ft - new_tri), 4 * num_triangles);

    delete[] vtx;
    delete[] tri;
    vtx = new_vtx;
    tri = new_tri;
    num_vertices += num_triangles + (num_triangles >> 1);
    num_triangles *= 4;
}

int Moon::findVertex(const Vector& v, const Vector* p, const Vector* end)
{
    /* helper function for Moon::subdivide() */

    int c = 0;

    for (; p < end; ++p, ++c) if (v == *p) return c;
    return -1;
}

int Moon::update()
{
    if (slerp < 1.) slerp += slerpspeed;

    /*if (key[KEY_SPACE])
    {
        if (!pressed)
        {
            Vector dir(sin(angle), -cos(angle), 0.);

            // screen direction to world space direction:
            QUAT y_quat, x_quat, quat;
            get_x_rotate_quat(&x_quat, -phi*(128/PI));
            get_y_rotate_quat(&y_quat, -theta*(128/PI));
            quat_mul(&x_quat, &y_quat, &quat);
            apply_quat(&quat, dir.x(), dir.y(), dir.z(), dir.adrx(), dir.adry(), dir.adrz());

            parts->kaboom(4, 50, align, 0.15*dir, -0.01, 0.01, makecol(255, 255, 0), 900, 8);
            pressed = true;
        }
    }
    else pressed = false;*/

    parts->update();

    /*Vector player(0., 0., -1.);
    QUAT quat;
    if (slerp >= 1.) quat = beginquat = endquat;
    else quat_interpolate(&beginquat, &endquat, slerp, &quat);
    apply_quat(&quat, player.x(), player.y(), player.z(),
            player.adrx(), player.adry(), player.adrz());*/

    int c = monsters.update(align, change);
    if (c != -1) return c;
    change = false;

    if (key[KEY_F11])
    {
        if (!wireframe_pressed) wireframe = !wireframe;
        wireframe_pressed=true;
    }
    else wireframe_pressed=false;
    return -1;
}

void Moon::draw(BITMAP* bmp)
{
    if (!selection_buffer || bmp->w != selection_buffer->w || bmp->h != selection_buffer->h)
    {
        destroy_bitmap(selection_buffer);
        selection_buffer = create_bitmap_ex(16, bmp->w, bmp->h);
        if (!selection_buffer)
            exitmsg("Not enough memory for selection buffer.");
    }

    QUAT quat;
    MATRIX_f matrix;

    if (slerp >= 1.) quat = beginquat = endquat;
    else quat_interpolate(&beginquat, &endquat, slerp, &quat);
    quat_to_matrix(&quat, &matrix);
    qtranslate_matrix_f(&matrix, 0., 0., 2.);

    /*MATRIX_f y_matrix, x_matrix, matrix;

    get_y_rotate_matrix_f(&y_matrix, theta*(128/PI));
    get_x_rotate_matrix_f(&x_matrix, phi*(128/PI));
    matrix_mul_f(&y_matrix, &x_matrix, &matrix);
    //matrix = identity_matrix_f;
    qtranslate_matrix_f(&matrix, 0., 0., 2.);
    matrix.t[2] = 2.;*/
    set_projection_viewport(0, (bmp->h - bmp->w)/2, bmp->w, bmp->w);

    // draw mars
    static Vector marsloc(100., 0., 0.);
    Vector rotated_marsloc;
    apply_matrix_f(&matrix, marsloc.x(), marsloc.y(), marsloc.z(),
            rotated_marsloc.adrx(), rotated_marsloc.adry(), rotated_marsloc.adrz());
    if (rotated_marsloc.z() > 0.)
    {
        persp_project_f(rotated_marsloc.x(), rotated_marsloc.y(), rotated_marsloc.z(),
                rotated_marsloc.adrx(), rotated_marsloc.adry());

        int x = (int) rotated_marsloc.x();
        int y = (int) rotated_marsloc.y();

        blit(mars, bmp, 0, 0, x, y, mars->w, mars->h);
    }

    // Project
    Vector* pend = vtx + num_vertices;
    Vector* q = rotated_vtx;
    Vector* p;
    for (p = vtx; p < pend; ++p, ++q)
        apply_matrix_f(&matrix, p->x(), p->y(), p->z(),
                q->adrx(), q->adry(), q->adrz());

    Vector* nm;
    Triangle* t;
    Triangle* tend;

    if (!wireframe)
    {
        tend = tri + num_triangles;
        nm=tri_normal;
        for (t = tri; t < tend; ++t, ++nm)
        {
            *nm = (rotated_vtx[t->i2]-rotated_vtx[t->i1]) |
                    (rotated_vtx[t->i3]-rotated_vtx[t->i1]);
        }
    }

    pend = vtx + num_vertices;
    q = rotated_vtx;
    for (p = vtx; p < pend; ++p, ++q)
        persp_project_f(q->x(), q->y(), q->z(), q->adrx(), q->adry());

    // Add edges to zsorter.
    memset((char*) edge_done, 0, num_vertices*num_vertices*sizeof(bool));
    zsorter_ptr = 0;

    nm = tri_normal;
    tend = tri + num_triangles;
    for (t = tri; t < tend; ++t, ++nm)
    {
        if (wireframe)
        {
            addEdge(t->i1, t->i2);
            addEdge(t->i2, t->i3);
            addEdge(t->i3, t->i1);
        }
        else
            if (nm->z() > 0.)
            {
                int c = (int) (191. * nm->cosz());
                addTriangle(t, makecol(c, c, c));
            }
    }

    monsters.draw(bmp, &matrix, this);

    // Perform painters' algorithm.
    qsort((void*) zsorter, zsorter_ptr, sizeof(ZSorter), zsorter_cmp);

    ZSorter* zend = zsorter + zsorter_ptr;
    for (ZSorter* z = zsorter; z < zend; ++z)
    {
        if (z->type == ZSorter::EDGE)
            aa_line(bmp, z->edge.x1, z->edge.y1, z->edge.x2, z->edge.y2, z->edge.c);
        else if (z->type == ZSorter::TRIANGLE)
        {
            int x1 = (int) rotated_vtx[z->tri.i1].x();
            int y1 = (int) rotated_vtx[z->tri.i1].y();
            int x2 = (int) rotated_vtx[z->tri.i2].x();
            int y2 = (int) rotated_vtx[z->tri.i2].y();
            int x3 = (int) rotated_vtx[z->tri.i3].x();
            int y3 = (int) rotated_vtx[z->tri.i3].y();

            triangle(bmp, x1, y1, x2, y2, x3, y3, z->tri.c);
        }
        else if (z->type == ZSorter::BITMAP)
        {
            int w = (z->bmp.p->w<<8) / z->z;
            int h = (z->bmp.p->h<<8) / z->z;

            //TRACE("%d, %d\n", w, h);

            if (w > 0 && h > 0)
                stretch_sprite(bmp, z->bmp.p, z->bmp.x-w/2, z->bmp.y-h/2, w, h);
        }
        else if (z->type == ZSorter::BITMAP_ADDITIVE)
        {
            if (worm_type) set_trans_blender(0, 0, 0, 63);
            else set_my_add_blender(0,0,0,255);
            draw_trans_sprite(bmp, z->bmp.p, z->bmp.x-z->bmp.p->w/2, z->bmp.y-z->bmp.p->h/2);
        }
    }

    parts->draw(bmp, &matrix);

    /*int n = 0;
    if (slerp < 1.) n = (int) (slerp * 20) & 1;
    else angle = atan2(mouse_y-bmp->h/2, mouse_x-bmp->w/2)+PI/2;
    pivot_sprite(bmp, sorcerer[n], bmp->w/2, bmp->h/2, sorcerer[n]->w/2, sorcerer[n]->h/2, ftofix(angle*(128/PI)));*/
    text_mode(0);
    textprintf(bmp, font, 0, 10, makecol(255, 255, 255), "%d Z-sorted", zsorter_ptr);
}

void Moon::addEdge(int i1, int i2)
{
    // helper for Moon::draw()
    if (zsorter_ptr < zsorter_size)
    {
        int a = MIN(i1, i2);
        int b = MAX(i1, i2);

        if (!edge_done[num_vertices*a+b])
        {
            edge_done[num_vertices*a+b] = true;
            zsorter[zsorter_ptr].type = ZSorter::EDGE;
            zsorter[zsorter_ptr].z = ((int) (256*rotated_vtx[a].z()) + (int) (256*rotated_vtx[b].z())) >> 1;
            zsorter[zsorter_ptr].edge.x1 = (int) rotated_vtx[a].x();
            zsorter[zsorter_ptr].edge.y1 = (int) rotated_vtx[a].y();
            zsorter[zsorter_ptr].edge.x2 = (int) rotated_vtx[b].x();
            zsorter[zsorter_ptr].edge.y2 = (int) rotated_vtx[b].y();
            int c = 255 - ((zsorter[zsorter_ptr].z - FOG_BEGIN) * 255 / (FOG_END-FOG_BEGIN));
            if (c > 0)
            {
                if (c > 255) c = 255;
                zsorter[zsorter_ptr].edge.c = makecol(c, c, c);
                ++zsorter_ptr;
            }
        }
    }
}

void Moon::addTriangle(Triangle* t, int c)
{
    if (zsorter_ptr < zsorter_size)
    {
        zsorter[zsorter_ptr].type = ZSorter::TRIANGLE;
        zsorter[zsorter_ptr].z = ((int) (256*rotated_vtx[t->i1].z()) +
                (int) (256*rotated_vtx[t->i2].z()) +
                (int) (256*rotated_vtx[t->i3].z())) / 3;
        zsorter[zsorter_ptr].tri.i1 = t->i1;
        zsorter[zsorter_ptr].tri.i2 = t->i2;
        zsorter[zsorter_ptr].tri.i3 = t->i3;
        zsorter[zsorter_ptr].tri.c = c;
        ++zsorter_ptr;
    }
}

void Moon::addBitmap(BITMAP* p, int x, int y, int z, int alpha)
{
    if (zsorter_ptr < zsorter_size)
    {
        zsorter[zsorter_ptr].type = alpha == -1 ? ZSorter::BITMAP : ZSorter::BITMAP_ADDITIVE;
        zsorter[zsorter_ptr].z = z;
        zsorter[zsorter_ptr].bmp.x = x;
        zsorter[zsorter_ptr].bmp.y = y;
        zsorter[zsorter_ptr].bmp.p = p;
        ++zsorter_ptr;
    }
}

int Moon::zsorter_cmp(const void* e1, const void* e2)
{
    const ZSorter* z1 = (const ZSorter*) e1;
    const ZSorter* z2 = (const ZSorter*) e2;

    return z2->z - z1->z;
}

void Moon::click(int x, int y)
{
    ASSERT(selection_buffer);

    // Add triangles to zsorter
    zsorter_ptr = 0;

    Triangle* tend = tri + num_triangles;
    for (Triangle* t = tri; t < tend; ++t) addTriangle(t, 1 + t - tri);

    // Perform painters' algorithm.
    qsort((void*) zsorter, zsorter_ptr, sizeof(ZSorter), zsorter_cmp);
    clear_bitmap(selection_buffer);

    ZSorter* zend = zsorter + zsorter_ptr;
    for (ZSorter* z = zsorter; z < zend; ++z)
    {
        ASSERT(z->type == ZSorter::TRIANGLE);

        int x1 = (int) rotated_vtx[z->tri.i1].x();
        int y1 = (int) rotated_vtx[z->tri.i1].y();
        int x2 = (int) rotated_vtx[z->tri.i2].x();
        int y2 = (int) rotated_vtx[z->tri.i2].y();
        int x3 = (int) rotated_vtx[z->tri.i3].x();
        int y3 = (int) rotated_vtx[z->tri.i3].y();

        triangle(selection_buffer, x1, y1, x2, y2, x3, y3, z->tri.c);
    }

    int c = getpixel(selection_buffer, x, y) - 1;
    if (c >= 0)
    {
        align = (vtx[tri[c].i1] + vtx[tri[c].i2] + vtx[tri[c].i3]) * (1./3);

        parts->kaboom(4, 1, align, 0., 0., 0., makecol(255, 0, 0), 1023, 16);
        //parts->kaboom(4, 2, align, 0.05*align, -0.01, 0.01, makecol(127, 127, 255), 1023, 8);

        Vector align2(align.z(), align.x(), align.y());

        theta = align2.theta()-PI;
        phi = align2.phi();//-PI*1.5;

        if (theta < 0.) theta += 2. * PI;
        if (phi < 0.) phi += 2. * PI;

        //TRACE("{ %g, %g, %g }\n", align.x(), align.y(), align.z());
        TRACE("theta=%g phi=%g\n", theta*180/PI, phi*180/PI);

        QUAT y_quat, x_quat;

        if (slerp < 1.)
        {
            QUAT quat;
            quat_interpolate(&beginquat, &endquat, slerp, &quat);
            beginquat = quat;
        }
        else beginquat = endquat;
        get_y_rotate_quat(&y_quat, theta*(128/PI));
        get_x_rotate_quat(&x_quat, phi*(128/PI));
        quat_mul(&y_quat, &x_quat, &endquat);
        slerp = 0.;
        slerpspeed = 2./sqrt((x-selection_buffer->w/2)*(x-selection_buffer->w/2) + (y-selection_buffer->h/2)*(y-selection_buffer->h/2));
        angle = atan2(y - selection_buffer->h/2, x - selection_buffer->w/2) + PI/2;
        change=true;
        //Vector up = align | Vector(0., 1., 0.);
        //get_align_matrix_f(&matrix, align.x(), align.y(), align.z(), up.x(), up.y(), up.z());

        //theta = align.theta();
        //phi = align.phi();
        //MATRIX_f matrix;
        //get_align_matrix_f(&matrix, -align.z(), align.y(), align.x(), 0., 1., 0.);
        //matrix_to_quat(&matrix, &quat);
    }
}
