/* Author: Tobi Vollebregt */

/*  Marsworm -- A small game written for the TINS Is Not SpeedHack 2003
 *  Copyright (C) 2003  Tobi Vollebregt
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *  See `License.txt', which contains a verbatim copy of the
 *  GNU General Public License, for details.
 *
 *  Please send your reaction to: tobivollebregt@hotmail.com
 */

#ifndef __tjv_moon_h__
#define __tjv_moon_h__

#include "vector.h"
#include "particle.h"
#include "monster.h"

struct BITMAP;

/* Implements the wireframe (and solid) sphere rendering
and the mapping of points onto the sphere. */

/* Actually this became a kind of "Game" or "God" class:
ie. everything, the particle engine, the worm, is owned by this class*/

class Moon
{
 public:

    Moon(): vtx(0), rotated_vtx(0), tri(0), edge_done(0), zsorter(0),
            selection_buffer(0),tri_normal(0) {}
    ~Moon();

    void initialize(int recur);

    int update();
    void draw(BITMAP* bmp);
    void click(int x, int y);

    void addBitmap(BITMAP* p, int x, int y, int z, int alpha=-1);

    int getScore() const { return monsters.getScore(); }

 private:

    struct Triangle
    {
        short i1, i2, i3;
        void setIndices(int a, int b, int c) { i1=a; i2=b; i3=c; }
        template<class T>
        void setIndices(T* p) { i1=p[0]; i2=p[1]; i3=p[2]; }
    };

    struct ZSorter
    {
        enum { EDGE, TRIANGLE, BITMAP, BITMAP_ADDITIVE };
        int z;
        int type;
        union {
            struct {
                short x1, y1, x2, y2;
                int c;
            } edge;
            struct {
                short i1, i2, i3;
                int c;
            } tri;
            struct {
                struct BITMAP* p;
                short x, y, alpha;
            } bmp;
        };
    };

    // subdivide the sphere 1 level
    void subdivide();
    int findVertex(const Vector&, const Vector*, const Vector*);

    void addEdge(int i1, int i2);
    void addTriangle(Triangle* tri, int c);
    static int zsorter_cmp(const void* e1, const void* e2);

    int num_vertices, num_triangles;
    Vector* vtx;
    Vector* rotated_vtx;
    Triangle* tri;
    float theta, phi, angle;
    bool* edge_done;
    ZSorter* zsorter;
    int zsorter_ptr, zsorter_size;
    BITMAP* selection_buffer;
    //MATRIX_f matrix;
    QUAT beginquat, endquat;
    float slerp,slerpspeed;
    BITMAP* mars;
    //BITMAP* sorcerer[2];
    Particle* parts;
    Vector align;
    //bool pressed;
    Monsters monsters;
    Vector* tri_normal;
    bool change;
    bool wireframe;
    bool wireframe_pressed;
};

#endif // !defined(__tjv_moon_h__)
