/* Author: Tobi Vollebregt */

/*  Marsworm -- A small game written for the TINS Is Not SpeedHack 2003
 *  Copyright (C) 2003  Tobi Vollebregt
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *  See `License.txt', which contains a verbatim copy of the
 *  GNU General Public License, for details.
 *
 *  Please send your reaction to: tobivollebregt@hotmail.com
 */

#include <allegro.h>
#include <allegro/internal/aintern.h>
#include <string.h>
#include "particle.h"

#define GRAVITY 0.02

Particle::Particle()
{
    memset((char*) parts, 0, sizeof(parts));
 
}

void Particle::update()
{
    Part* pend = parts + NUM_PARTS;
    for (Part* p = parts; p < pend; ++p)
        if (p->alpha > 0)
        {
            p->pos += p->speed;
            if (p->pos.magnitude() < 1) { p->pos.normalize(); p->speed = -0.5 * p->speed; }
            p->speed -= GRAVITY * p->pos * (1./ (p->pos * p->pos));
            //p->speed *= 0.99;
            p->alpha-=p->alphaspeed;
        }
}

void Particle::draw(BITMAP* bmp, MATRIX_f* matrix)
{
    drawing_mode(DRAW_MODE_TRANS, 0, 0, 0);
    set_trans_blender(0, 0, 0, 0);

    Part* pend = parts + NUM_PARTS;
    for (Part* p = parts; p < pend; ++p) if (p->alpha > 0)
    {
        float fx, fy, fz;

        apply_matrix_f(matrix, p->pos.x(), p->pos.y(), p->pos.z(), &fx, &fy, &fz);

        _blender_alpha = p->alpha >> 2;

        if (fz > 0.)
        {
            persp_project_f(fx, fy, fz, &fx, &fy);

            int ix = (int) fx, iy = (int) fy;

            circlefill(bmp, ix, iy, (int) (p->radius / fz), p->color);
        }
    }
    solid_mode();
}

void Particle::kaboom(float radius, int n, const Vector& pos, const Vector& speed, float r1, float r2, int color, int alpha, int alphaspeed)
{
    Part* pend = parts + NUM_PARTS;
    for (Part* p = parts; p < pend; ++p)
        if (p->alpha <= 0)
        {
            p->radius = radius;
            p->pos = pos;
            p->speed = speed + Vector((float) rand() / RAND_MAX * (r2-r1) + r1,
                    (float) rand() / RAND_MAX * (r2-r1) + r1,
                    (float) rand() / RAND_MAX * (r2-r1) + r1);
            p->color = color;
            p->alpha = alpha;
            p->alphaspeed = alphaspeed;
            if (--n == 0) break;
        }
}
