/* Author: Tobi Vollebregt */

/*  Marsworm -- A small game written for the TINS Is Not SpeedHack 2003
 *  Copyright (C) 2003  Tobi Vollebregt
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *  See `License.txt', which contains a verbatim copy of the
 *  GNU General Public License, for details.
 *
 *  Please send your reaction to: tobivollebregt@hotmail.com
 */

#include "vector.h"

#define EPSILON (.001)

Vector Vector::Polar(value_type theta, value_type phi, value_type r)
{
    return Vector(
        cos(phi) * cos(theta) * r,
        cos(phi) * sin(theta) * r,
        sin(phi) * r);
}

bool Vector::operator==(const Vector& w) const
{
    if (fabs(x_ - w.x()) <= EPSILON &&
        fabs(y_ - w.y()) <= EPSILON &&
        fabs(z_ - w.z()) <= EPSILON) return true;
    return false;
}

void Vector::setPolar(value_type theta, value_type phi, value_type r)
{
    x_ = cos(phi) * cos(theta) * r;
    y_ = cos(phi) * sin(theta) * r;
    z_ = sin(phi) * r;
}

void Vector::clamp(value_type lo, value_type hi)
{
    if (x_ < lo) x_ = lo;
    if (x_ > hi) x_ = hi;
    if (y_ < lo) y_ = lo;
    if (y_ > hi) y_ = hi;
    if (z_ < lo) z_ = lo;
    if (z_ > hi) z_ = hi;
}

bool pointInPoly(const Vector& pt, const Vector* ptr, int size)
{
    const Vector* endptr = ptr + size;
    Vector prev_r = ptr[size-1] - pt;
    Vector::value_type prev_mag = prev_r.magnitude(), x = 0.;

    for (; ptr != endptr; ++ptr)
    {
        Vector r = *ptr - pt;
        Vector::value_type mag = r.magnitude();

        x += asin(r * prev_r / (mag * prev_mag));

        prev_r = r;
        prev_mag = mag;
    }

    if (x > EPSILON) return false;
    return true;
}
